@@ -39,10 +39,13 @@ class Client:
     def __init__(self, url, traversal_source, protocol_factory=None,
                  transport_factory=None, pool_size=None, max_workers=None,
                  message_serializer=None, username="", password="",
-                 headers=None, session=""):
+                 kerberized_service="", headers=None, session="",
+                 **transport_kwargs):
         self._url = url
         self._headers = headers
         self._traversal_source = traversal_source
+        if "max_content_length" not in transport_kwargs:
+            transport_kwargs["max_content_length"] = 10 * 1024 * 1024
         if message_serializer is None:
             message_serializer = serializer.GraphSONSerializersV3d0()
         self._message_serializer = message_serializer
@@ -52,19 +55,21 @@ class Client:
         self._sessionEnabled = (session != "")
         if transport_factory is None:
             try:
-                from gremlin_python.driver.tornado.transport import (
-                    TornadoTransport)
+                from gremlin_python.driver.aiohttp.transport import (
+                    AiohttpTransport)
             except ImportError:
-                raise Exception("Please install Tornado or pass"
+                raise Exception("Please install AIOHTTP or pass "
                                 "custom transport factory")
             else:
-                transport_factory = lambda: TornadoTransport()
+                def transport_factory():
+                    return AiohttpTransport(**transport_kwargs)
         self._transport_factory = transport_factory
         if protocol_factory is None:
-            protocol_factory = lambda: protocol.GremlinServerWSProtocol(
+            def protocol_factory(): return protocol.GremlinServerWSProtocol(
                 self._message_serializer,
                 username=self._username,
-                password=self._password)
+                password=self._password,
+                kerberized_service=kerberized_service)
         self._protocol_factory = protocol_factory
         if self._sessionEnabled:
             if pool_size is None:
@@ -76,9 +81,9 @@ class Client:
         self._pool_size = pool_size
         # This is until concurrent.futures backport 3.1.0 release
         if max_workers is None:
-            # Use this number because ThreadPoolExecutor is often
-            # used to overlap I/O instead of CPU work.
-            max_workers = (cpu_count() or 1) * 5
+            # If your application is overlapping Gremlin I/O on multiple threads
+            # consider passing kwarg max_workers = (cpu_count() or 1) * 5
+            max_workers = pool_size
         self._executor = ThreadPoolExecutor(max_workers=max_workers)
         # Threadsafe queue
         self._pool = queue.Queue()
@@ -123,10 +128,10 @@ class Client:
             self._transport_factory, self._executor, self._pool,
             headers=self._headers)
 
-    def submit(self, message, bindings=None):
-        return self.submitAsync(message, bindings=bindings).result()
+    def submit(self, message, bindings=None, request_options=None):
+        return self.submitAsync(message, bindings=bindings, request_options=request_options).result()
 
-    def submitAsync(self, message, bindings=None):
+    def submitAsync(self, message, bindings=None, request_options=None):
         if isinstance(message, traversal.Bytecode):
             message = request.RequestMessage(
                 processor='traversal', op='bytecode',
@@ -143,4 +148,6 @@ class Client:
                 message = message._replace(processor='session')
                 message.args.update({'session': self._session})
         conn = self._pool.get(True)
+        if request_options:
+            message.args.update(request_options)
         return conn.write(message)
