@@ -1,22 +1,23 @@
-'''
-Licensed to the Apache Software Foundation (ASF) under one
-or more contributor license agreements.  See the NOTICE file
-distributed with this work for additional information
-regarding copyright ownership.  The ASF licenses this file
-to you under the Apache License, Version 2.0 (the
-"License"); you may not use this file except in compliance
-with the License.  You may obtain a copy of the License at
-
-http://www.apache.org/licenses/LICENSE-2.0
-
-Unless required by applicable law or agreed to in writing,
-software distributed under the License is distributed on an
-"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
-KIND, either express or implied.  See the License for the
-specific language governing permissions and limitations
-under the License.
-'''
+#
+# Licensed to the Apache Software Foundation (ASF) under one
+# or more contributor license agreements.  See the NOTICE file
+# distributed with this work for additional information
+# regarding copyright ownership.  The ASF licenses this file
+# to you under the Apache License, Version 2.0 (the
+# "License"); you may not use this file except in compliance
+# with the License.  You may obtain a copy of the License at
+#
+# http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing,
+# software distributed under the License is distributed on an
+# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+# KIND, either express or implied.  See the License for the
+# specific language governing permissions and limitations
+# under the License.
+#
 
+import copy
 from aenum import Enum
 from .. import statics
 from ..statics import long
@@ -26,7 +27,6 @@ class Traversal(object):
         self.graph = graph
         self.traversal_strategies = traversal_strategies
         self.bytecode = bytecode
-        self.side_effects = TraversalSideEffects()
         self.traversers = None
         self.last_traverser = None
 
@@ -60,6 +60,7 @@ class Traversal(object):
         return set(iter(self))
 
     def iterate(self):
+        self.bytecode.add_step("none")
         while True:
             try: self.nextTraverser()
             except StopIteration: return self
@@ -106,7 +107,6 @@ class Traversal(object):
                 future.set_exception(e)
             else:
                 self.traversers = iter(traversal.traversers)
-                self.side_effects = traversal.side_effects
                 if cb:
                     try:
                         result = cb(self)
@@ -141,63 +141,67 @@ statics.add_static('OUT', Direction.OUT)
 statics.add_static('IN', Direction.IN)
 statics.add_static('BOTH', Direction.BOTH)
 
-GraphSONVersion = Enum('GraphSONVersion', ' V1_0 V2_0')
+GraphSONVersion = Enum('GraphSONVersion', ' V1_0 V2_0 V3_0')
 
 statics.add_static('V1_0', GraphSONVersion.V1_0)
 statics.add_static('V2_0', GraphSONVersion.V2_0)
+statics.add_static('V3_0', GraphSONVersion.V3_0)
 
-GryoVersion = Enum('GryoVersion', ' V1_0')
+GryoVersion = Enum('GryoVersion', ' V1_0 V3_0')
 
 statics.add_static('V1_0', GryoVersion.V1_0)
+statics.add_static('V3_0', GryoVersion.V3_0)
 
-Operator = Enum('Operator', ' addAll and_ assign div max min minus mult or_ sum sumLong')
+Order = Enum('Order', ' asc desc shuffle')
 
-statics.add_static('sum', Operator.sum)
-statics.add_static('minus', Operator.minus)
-statics.add_static('mult', Operator.mult)
-statics.add_static('div', Operator.div)
-statics.add_static('min', Operator.min)
-statics.add_static('max', Operator.max)
-statics.add_static('assign', Operator.assign)
-statics.add_static('and_', Operator.and_)
-statics.add_static('or_', Operator.or_)
-statics.add_static('addAll', Operator.addAll)
-statics.add_static('sumLong', Operator.sumLong)
-
-Order = Enum('Order', ' decr incr keyDecr keyIncr shuffle valueDecr valueIncr')
-
-statics.add_static('incr', Order.incr)
-statics.add_static('decr', Order.decr)
-statics.add_static('keyIncr', Order.keyIncr)
-statics.add_static('valueIncr', Order.valueIncr)
-statics.add_static('keyDecr', Order.keyDecr)
-statics.add_static('valueDecr', Order.valueDecr)
 statics.add_static('shuffle', Order.shuffle)
+statics.add_static('asc', Order.asc)
+statics.add_static('desc', Order.desc)
 
 Pick = Enum('Pick', ' any none')
 
 statics.add_static('any', Pick.any)
 statics.add_static('none', Pick.none)
 
-Pop = Enum('Pop', ' all_ first last')
+Pop = Enum('Pop', ' all_ first last mixed')
 
 statics.add_static('first', Pop.first)
 statics.add_static('last', Pop.last)
 statics.add_static('all_', Pop.all_)
+statics.add_static('mixed', Pop.mixed)
 
 Scope = Enum('Scope', ' global_ local')
 
 statics.add_static('global_', Scope.global_)
 statics.add_static('local', Scope.local)
 
-T = Enum('T', ' id key label value')
 
-statics.add_static('label', T.label)
+T = Enum('T', ' id id_ key label value')
+
 statics.add_static('id', T.id)
+statics.add_static('label', T.label)
+statics.add_static('id_', T.id_)
 statics.add_static('key', T.key)
 statics.add_static('value', T.value)
 
 
+Operator = Enum('Operator', ' addAll and_ assign div max max_ min min_ minus mult or_ sum sum_ sumLong')
+
+statics.add_static('sum_', Operator.sum_)
+statics.add_static('sum', Operator.sum_)
+statics.add_static('minus', Operator.minus)
+statics.add_static('mult', Operator.mult)
+statics.add_static('div', Operator.div)
+statics.add_static('min', Operator.min_)
+statics.add_static('min_', Operator.min_)
+statics.add_static('max_', Operator.max_)
+statics.add_static('assign', Operator.assign)
+statics.add_static('and_', Operator.and_)
+statics.add_static('or_', Operator.or_)
+statics.add_static('addAll', Operator.addAll)
+statics.add_static('sumLong', Operator.sumLong)
+
+
 class P(object):
     def __init__(self, operator, value, other=None):
         self.operator = operator
@@ -250,11 +254,17 @@ class P(object):
 
     @staticmethod
     def within(*args):
-        return P("within", *args)
-
+        if len(args) == 1 and type(args[0]) == list:
+            return P("within", args[0])
+        else:
+            return P("within", list(args))
+        
     @staticmethod
     def without(*args):
-        return P("without", *args)
+        if len(args) == 1 and type(args[0]) == list:
+            return P("without", args[0])
+        else:
+            return P("without", list(args))
 
     def and_(self, arg):
         return P("and", self, arg)
@@ -268,63 +278,239 @@ class P(object):
     def __repr__(self):
         return self.operator + "(" + str(self.value) + ")" if self.other is None else self.operator + "(" + str(self.value) + "," + str(self.other) + ")"
 
+
 def between(*args):
     return P.between(*args)
-statics.add_static('between',between)
+
 
 def eq(*args):
     return P.eq(*args)
-statics.add_static('eq',eq)
+
 
 def gt(*args):
     return P.gt(*args)
-statics.add_static('gt',gt)
+
 
 def gte(*args):
     return P.gte(*args)
-statics.add_static('gte',gte)
+
 
 def inside(*args):
     return P.inside(*args)
-statics.add_static('inside',inside)
+
 
 def lt(*args):
     return P.lt(*args)
-statics.add_static('lt',lt)
+
 
 def lte(*args):
     return P.lte(*args)
-statics.add_static('lte',lte)
+
 
 def neq(*args):
     return P.neq(*args)
-statics.add_static('neq',neq)
+
 
 def not_(*args):
     return P.not_(*args)
-statics.add_static('not_',not_)
+
 
 def outside(*args):
     return P.outside(*args)
-statics.add_static('outside',outside)
 
-def test(*args):
-    return P.test(*args)
-statics.add_static('test',test)
 
 def within(*args):
     return P.within(*args)
-statics.add_static('within',within)
+
 
 def without(*args):
     return P.without(*args)
-statics.add_static('without',without)
+
+
+statics.add_static('between', between)
+
+statics.add_static('eq', eq)
+
+statics.add_static('gt', gt)
+
+statics.add_static('gte', gte)
+
+statics.add_static('inside', inside)
+
+statics.add_static('lt', lt)
+
+statics.add_static('lte', lte)
+
+statics.add_static('neq', neq)
+
+statics.add_static('not_', not_)
+
+statics.add_static('outside', outside)
+
+statics.add_static('within', within)
+
+statics.add_static('without', without)
+
+
+class TextP(P):
+    def __init__(self, operator, value, other=None):
+        P.__init__(self, operator, value, other)
+
+    @staticmethod
+    def containing(*args):
+        return TextP("containing", *args)
+
+    @staticmethod
+    def endingWith(*args):
+        return TextP("endingWith", *args)
+
+    @staticmethod
+    def notContaining(*args):
+        return TextP("notContaining", *args)
+
+    @staticmethod
+    def notEndingWith(*args):
+        return TextP("notEndingWith", *args)
+
+    @staticmethod
+    def notStartingWith(*args):
+        return TextP("notStartingWith", *args)
+
+    @staticmethod
+    def startingWith(*args):
+        return TextP("startingWith", *args)
+
+    def __eq__(self, other):
+        return isinstance(other, self.__class__) and self.operator == other.operator and self.value == other.value and self.other == other.other
+
+    def __repr__(self):
+        return self.operator + "(" + str(self.value) + ")" if self.other is None else self.operator + "(" + str(self.value) + "," + str(self.other) + ")"
+
+
+def containing(*args):
+    return TextP.containing(*args)
+
+
+def endingWith(*args):
+    return TextP.endingWith(*args)
+
+
+def notContaining(*args):
+    return TextP.notContaining(*args)
+
+
+def notEndingWith(*args):
+    return TextP.notEndingWith(*args)
+
+
+def notStartingWith(*args):
+    return TextP.notStartingWith(*args)
+
+
+def startingWith(*args):
+    return TextP.startingWith(*args)
+
+
+statics.add_static('containing', containing)
+
+statics.add_static('endingWith', endingWith)
+
+statics.add_static('notContaining', notContaining)
+
+statics.add_static('notEndingWith', notEndingWith)
+
+statics.add_static('notStartingWith', notStartingWith)
+
+statics.add_static('startingWith', startingWith)
+
+
+
+
+'''
+IO
+'''
+
+
+class IO(object):
+
+    graphml = "graphml"
+
+    graphson = "graphson"
+
+    gryo = "gryo"
+
+    reader = "~tinkerpop.io.reader"
+
+    registry = "~tinkerpop.io.registry"
+
+    writer = "~tinkerpop.io.writer"
+
+
+'''
+ConnectedComponent
+'''
+
+
+class ConnectedComponent(object):
+
+    component = "gremlin.connectedComponentVertexProgram.component"
+
+    edges = "~tinkerpop.connectedComponent.edges"
+
+    propertyName = "~tinkerpop.connectedComponent.propertyName"
+
+
+'''
+ShortestPath
+'''
+
+
+class ShortestPath(object):
+
+    distance = "~tinkerpop.shortestPath.distance"
+
+    edges = "~tinkerpop.shortestPath.edges"
+
+    includeEdges = "~tinkerpop.shortestPath.includeEdges"
+
+    maxDistance = "~tinkerpop.shortestPath.maxDistance"
+
+    target = "~tinkerpop.shortestPath.target"
+
+
+'''
+PageRank
+'''
+
+
+class PageRank(object):
+
+    edges = "~tinkerpop.pageRank.edges"
+
+    propertyName = "~tinkerpop.pageRank.propertyName"
+
+    times = "~tinkerpop.pageRank.times"
+
+
+'''
+PeerPressure
+'''
+
+
+class PeerPressure(object):
+
+    edges = "~tinkerpop.peerPressure.edges"
+
+    propertyName = "~tinkerpop.peerPressure.propertyName"
+
+    times = "~tinkerpop.peerPressure.times"
 
 
 '''
 TRAVERSER
 '''
 
+
 class Traverser(object):
     def __init__(self, object, bulk=None):
         if bulk is None:
@@ -338,31 +524,17 @@ class Traverser(object):
     def __eq__(self, other):
         return isinstance(other, self.__class__) and self.object == other.object
 
-'''
-TRAVERSAL SIDE-EFFECTS
-'''
-
-class TraversalSideEffects(object):
-    def keys(self):
-        return set()
-
-    def get(self, key):
-        raise KeyError(key)
-
-    def __getitem__(self, key):
-        return self.get(key)
-
-    def __repr__(self):
-        return "sideEffects[size:" + str(len(self.keys())) + "]"
 
 '''
 TRAVERSAL STRATEGIES
 '''
 
+
 class TraversalStrategies(object):
     global_cache = {}
+
     def __init__(self, traversal_strategies=None):
-        self.traversal_strategies = traversal_strategies.traversal_strategies if traversal_strategies is not None else []
+        self.traversal_strategies =             traversal_strategies.traversal_strategies if traversal_strategies is not None else []
 
     def add_strategies(self, traversal_strategies):
         self.traversal_strategies = self.traversal_strategies + traversal_strategies
@@ -380,7 +552,8 @@ class TraversalStrategies(object):
 
 
 class TraversalStrategy(object):
-    def __init__(self, strategy_name=None, configuration=None):
+    def __init__(self, strategy_name=None, configuration=None, fqcn=None):
+        self.fqcn = fqcn
         self.strategy_name = type(self).__name__ if strategy_name is None else strategy_name
         self.configuration = {} if configuration is None else configuration
 
@@ -399,10 +572,12 @@ class TraversalStrategy(object):
     def __repr__(self):
         return self.strategy_name
 
+
 '''
 BYTECODE
 '''
 
+
 class Bytecode(object):
     def __init__(self, bytecode=None):
         self.source_instructions = []
@@ -430,8 +605,24 @@ class Bytecode(object):
         else:
             return False
 
+    def __copy__(self):
+        bb = Bytecode()
+        bb.source_instructions = self.source_instructions
+        bb.step_instructions = self.step_instructions
+        bb.bindings = self.bindings
+        return bb
+
+    def __deepcopy__(self, memo={}):
+        bb = Bytecode()
+        bb.source_instructions = copy.deepcopy(self.source_instructions, memo)
+        bb.step_instructions = copy.deepcopy(self.step_instructions, memo)
+        bb.bindings = copy.deepcopy(self.bindings, memo)
+        return bb
+
     def __convertArgument(self,arg):
         if isinstance(arg, Traversal):
+            if arg.graph is not None:
+                raise TypeError("The child traversal of " + arg + " was not spawned anonymously - use the __ class rather than a TraversalSource to construct the child traversal")
             self.bindings.update(arg.bytecode.bindings)
             return arg.bytecode
         elif isinstance(arg, dict):
@@ -449,9 +640,9 @@ class Bytecode(object):
             for item in arg:
                 newSet.add(self.__convertArgument(item))
             return newSet
-        elif isinstance(arg, tuple) and 2 == len(arg) and isinstance(arg[0], str):
-            self.bindings[arg[0]] = arg[1]
-            return Binding(arg[0],self.__convertArgument(arg[1]))
+        elif isinstance(arg, Binding):
+            self.bindings[arg.key] = arg.value
+            return Binding(arg.key, self.__convertArgument(arg.value))
         else:
             return arg
 
@@ -464,14 +655,18 @@ class Bytecode(object):
 BINDINGS
 '''
 
+
 class Bindings(object):
-    def of(self,key,value):
+
+    @staticmethod
+    def of(key, value):
         if not isinstance(key, str):
             raise TypeError("Key must be str")
-        return (key,value)
+        return Binding(key, value)
+
 
 class Binding(object):
-    def __init__(self,key,value):
+    def __init__(self, key, value):
         self.key = key
         self.value = value
 
@@ -482,4 +677,33 @@ class Binding(object):
         return hash(self.key) + hash(self.value)
 
     def __repr__(self):
-        return "binding[" + self.key + "=" + str(self.value) + "]"
\ No newline at end of file
+        return "binding[" + self.key + "=" + str(self.value) + "]"
+
+
+'''
+WITH OPTIONS
+'''
+
+
+class WithOptions(object):
+
+    tokens = "~tinkerpop.valueMap.tokens"
+
+    none = 0
+
+    ids = 1
+
+    labels = 2
+
+    keys = 4
+
+    values = 8
+
+    all = 15
+
+    indexer = "~tinkerpop.index.indexer"
+
+    list = 0
+
+    map = 1
+
